import threading
from typing import Dict, Generic, Optional, TypeVar

ValueType = TypeVar("ValueType")


class PThreadStore(Generic[ValueType]):
    def __init__(self):
        self.items: Dict[int, ValueType] = {}

    def pthread_id(self) -> int:
        return threading.get_ident()

    def set(self, value: ValueType) -> None:
        self.items[self.pthread_id()] = value

    def get(self) -> Optional[ValueType]:
        return self.items.get(self.pthread_id())

    def clear(self) -> None:
        self.items.pop(self.pthread_id(), None)



import sys
from typing import Callable, Literal, Union

from abstra_internals.controllers.execution_stdio import (
    ExecutionStdioController,
    UnboundPthread,
)
from abstra_internals.env_masker import GLOBAL_MASKER
from abstra_internals.environment import IS_PRODUCTION
from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories import (
    execution_logs_repository,
    execution_repository,
)


class StdioPatcher:
    sys_stdout_write = sys.stdout.write
    sys_stderr_write = sys.stderr.write
    _stdio_controller: ExecutionStdioController
    _debug: bool = False

    @classmethod
    def apply(cls, debug: bool):
        cls._stdio_controller = ExecutionStdioController(
            execution_logs_repository=execution_logs_repository,
            execution_repository=execution_repository,
        )

        cls._debug = debug

        sys.stdout.write = lambda raw: cls._wrapper("stdout", cls.sys_stdout_write, raw)
        sys.stderr.write = lambda raw: cls._wrapper("stderr", cls.sys_stderr_write, raw)

    @classmethod
    def reset(cls):
        sys.stdout.write = cls.sys_stdout_write
        sys.stderr.write = cls.sys_stderr_write

    @classmethod
    def _wrapper(
        cls,
        type: Literal["stdout", "stderr"],
        sys_write: Callable,
        raw: Union[str, bytearray],
    ):
        text = raw.decode("utf-8") if not isinstance(raw, str) else raw

        if IS_PRODUCTION:
            text = GLOBAL_MASKER.mask(text)

        try:
            execution_id = cls._stdio_controller.write_stdio(type, text)
            short_id = execution_id.split(sep="-")[0]
            term_output = f"[RUN {short_id}] {text}" if text.strip() else text
            sys_write(term_output)
        except UnboundPthread:
            sys_write(text)
        except Exception as e:
            AbstraLogger.capture_exception(e)
            if cls._debug:
                sys_write(f"[STDIO_PATCHER]: {e}")
        finally:
            sys.stdout.flush()
            return len(text)



import copy
import json
import uuid
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Dict, List, Literal, Mapping, Optional, Union

import requests
from pydantic import ConfigDict, Field
from pydantic.dataclasses import dataclass

from abstra_internals.environment import SIDECAR_HEADERS, SIDECAR_URL
from abstra_internals.utils.datetime import from_utc_iso_string
from abstra_internals.utils.filter import FilterCondition, evaluate_filter_condition
from abstra_internals.utils.validate import validate_json

Status = Literal["waiting", "running", "finished", "failed", "abandoned"]

end_status = ["failed", "finished", "abandoned"]

end_status: List[Status] = ["failed", "finished", "abandoned"]

status_transitions: Dict[Status, List[Status]] = {
    "waiting": ["running"],
    "running": end_status,
}


@dataclass
class Pagination:
    limit: int
    offset: int

    @staticmethod
    def from_dict(data: dict) -> "Pagination":
        return Pagination(
            limit=data.get("limit", 10),
            offset=data.get("offset", 0),
        )

    def to_dict(self) -> dict:
        return dict((k, v) for k, v in self.__dict__.items() if v is not None)


@dataclass
class PaginatedListResponse:
    stage_runs: List["StageRun"]
    total_count: int

    @staticmethod
    def from_dict(data: dict) -> "PaginatedListResponse":
        return PaginatedListResponse(
            stage_runs=[StageRun.from_dict(dto) for dto in data.get("stageRuns", [])],
            total_count=data.get("totalCount", 0),
        )

    def to_dict(self) -> dict:
        return dict((k, v) for k, v in self.__dict__.items() if v is not None)


@dataclass
class GetStageRunByQueryFilter:
    stage: Optional[Union[str, List[str]]] = None
    assignee: Optional[str] = None
    parent_id: Optional[str] = None
    status: Optional[List[Status]] = None
    data: Optional[Dict] = None
    search: Optional[str] = None
    data_conditions: Optional[FilterCondition] = None

    @staticmethod
    def from_dict(data: dict) -> "GetStageRunByQueryFilter":
        return GetStageRunByQueryFilter(
            stage=data.get("stage", None),
            assignee=data.get("assignee", None),
            parent_id=data.get("parent_id", None),
            status=data.get("status", None),
            data=data.get("data", None),
            search=data.get("search", None),
            data_conditions=FilterCondition.from_dict(data.get("data_conditions", {})),
        )

    def to_dict(self) -> dict:
        return dict((k, v) for k, v in self.__dict__.items() if v is not None)


@dataclass(config=ConfigDict(populate_by_name=True))
class StageRun:
    id: str
    stage: str
    data: dict
    status: Status
    created_at: datetime = Field(serialization_alias="createdAt")
    updated_at: datetime = Field(default=None, serialization_alias="updatedAt")
    parent_id: Optional[str] = Field(default=None, serialization_alias="parentId")
    execution_id: Optional[str] = Field(default=None, serialization_alias="executionId")

    @staticmethod
    def from_dict(data: dict) -> "StageRun":
        return StageRun(
            id=data["id"],
            stage=data["stage"],
            data=data["data"],
            status=data["status"],
            parent_id=data.get("parentId"),
            execution_id=data.get("executionId"),
            created_at=from_utc_iso_string(data["createdAt"]),
            updated_at=from_utc_iso_string(data["updatedAt"]),
        )

    def __getitem__(self, key):
        return self.data.get(key)

    def __setitem__(self, key, value):
        self.data[key] = value

    def get(self, key, default=None):
        return self.data.get(key, default)

    @property
    def is_end(self):
        return self.status in end_status

    @property
    def is_abandoned(self):
        return self.status == "abandoned"

    def to_dto(self):
        return {
            "id": self.id,
            "stage": self.stage,
            "data": self.data,
            "status": self.status,
            "parentId": self.parent_id,
            "executionId": self.execution_id,
            "createdAt": self.created_at.isoformat(),
            "updatedAt": self.updated_at.isoformat(),
        }

    def clone(self):
        return StageRun(**copy.deepcopy(self.__dict__))

    def clone_to_waiting(self):
        new_stage_run = self.clone()
        new_stage_run.id = str(uuid.uuid4())
        new_stage_run.status = "waiting"
        new_stage_run.created_at = datetime.now()
        new_stage_run.updated_at = datetime.now()
        return new_stage_run


class DetachedStageRun(StageRun):
    pass


class StageRunRepository(ABC):
    valid_filter_keys = ["data", "stage", "status", "parent_id"]

    def validate_filter_keys(self, filter: Dict) -> None:
        for key in filter.keys():
            if key not in self.valid_filter_keys:
                raise Exception(f"Invalid filter key {key}")

    def hydrate_next_dto(self, parent: StageRun, dtos: List[Dict]) -> List[Dict]:
        next_dtos = []

        for dto in dtos:
            stage = dto.get("stage")
            data = {**parent.data, **dto.get("data", {})}
            next_dtos.append(dict(stage=stage, data=data))

        return next_dtos

    @abstractmethod
    def find(
        self, filter: GetStageRunByQueryFilter, filter_detached: Optional[bool] = True
    ) -> List[StageRun]:
        raise NotImplementedError()

    @abstractmethod
    def find_ancestors(self, id: str) -> List[StageRun]:
        raise NotImplementedError()

    @abstractmethod
    def find_leaves(
        self,
        filter: GetStageRunByQueryFilter,
        pagination: Pagination,
    ) -> PaginatedListResponse:
        raise NotImplementedError()

    @abstractmethod
    def find_past(
        self, filter: GetStageRunByQueryFilter, pagination: Pagination
    ) -> PaginatedListResponse:
        raise NotImplementedError()

    @abstractmethod
    def get(self, id: str) -> StageRun:
        raise NotImplementedError()

    @abstractmethod
    def change_status(
        self, id: str, status: Status, execution_id: Optional[str] = None
    ) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def create_initial(self, stage: str, data: dict = {}) -> StageRun:
        raise NotImplementedError()

    @abstractmethod
    def create_detached(self, stage_id: str, thread_data: dict = {}) -> StageRun:
        raise NotImplementedError()

    @abstractmethod
    def remove(self, id: str) -> None:
        raise NotImplementedError()

    @abstractmethod
    def create_next(self, parent: StageRun, dtos: List[Dict]) -> List[StageRun]:
        raise NotImplementedError()

    @abstractmethod
    def fork(
        self,
        stage_run: StageRun,
        parent_is_iterator: bool,
        custom_thread_data: Optional[str],
    ) -> StageRun:
        raise NotImplementedError()

    @abstractmethod
    def retry(self, stage_run: StageRun) -> StageRun:
        raise NotImplementedError()

    @abstractmethod
    def update_data(self, stage_run_id: str, data: dict) -> bool:
        raise NotImplementedError()

    def get_not_abandoned_stage_run(
        self, stage_run: StageRun, stage_id: str
    ) -> StageRun:
        if stage_run.status != "abandoned":
            return stage_run

        filter = GetStageRunByQueryFilter()
        filter.parent_id = stage_run.id
        filter.stage = stage_id

        next_matches = self.find(filter)

        assert (
            len(next_matches) == 1
        ), "Internal error: abandoned stage run does not have exactly one child"

        return self.get_not_abandoned_stage_run(next_matches[0], stage_id)

    def acquire_lock(self, stage_run_id: str, execution_id: str) -> bool:
        return self.change_status(stage_run_id, "running", execution_id)


class LocalStageRunRepository(StageRunRepository):
    _stage_runs: List[StageRun]

    def __init__(self):
        self._stage_runs = []
        self._locks = []

    def clear(self):
        self._stage_runs = []
        self._locks = []

    def get(self, id: str) -> StageRun:
        for stage_run in self._stage_runs:
            if stage_run.id == id:
                return stage_run

        raise Exception(f"StageRun with id {id} not found")

    def find(
        self, filter: GetStageRunByQueryFilter, filter_detached: Optional[bool] = True
    ) -> List[StageRun]:
        results: List[StageRun] = []

        for stage_run in self._stage_runs:
            if isinstance(stage_run, DetachedStageRun) and filter_detached:
                continue

            if filter.stage and stage_run.stage not in filter.stage:
                continue

            if filter.data and not self._compare_data(stage_run.data, filter.data):
                continue

            if filter.status and stage_run.status not in filter.status:
                continue

            if filter.parent_id and stage_run.parent_id != filter.parent_id:
                continue

            if filter.search and not self._compare_search(stage_run, filter.search):
                continue

            if filter.data_conditions and not evaluate_filter_condition(
                filter.data_conditions, stage_run.data
            ):
                continue

            results.append(stage_run)

        return results

    def remove(self, id: str) -> None:
        self._stage_runs = [
            stage_run for stage_run in self._stage_runs if stage_run.id != id
        ]

    def _compare_data(self, data: dict, filter: dict) -> bool:
        for key, value in filter.items():
            if key not in data or data[key] not in value:
                return False
        return True

    def _compare_search(self, stage_run: StageRun, search: str) -> bool:
        # TODO: filter by stage title and stage type
        status_match = search.lower() in stage_run.status.lower()
        data_match = False
        for key, value in stage_run.data.items():
            if search.lower() in value.lower():
                data_match = True
        return status_match or data_match

    def find_ancestors(self, id: str) -> List[StageRun]:
        ancestors = []
        stage_run = self.get(id)
        ancestors.append(stage_run)
        while stage_run.parent_id is not None:
            stage_run = self.get(stage_run.parent_id)
            ancestors.insert(0, stage_run)
        return ancestors

    def find_leaves(
        self,
        filter: GetStageRunByQueryFilter,
        pagination: Pagination,
    ) -> PaginatedListResponse:
        filter_matches = self.find(filter)
        parent_ids = set(u.parent_id for u in self._stage_runs)

        leaves = sorted(
            [
                stage_run
                for stage_run in filter_matches
                if stage_run.id not in parent_ids
            ],
            key=lambda x: x.created_at,
            reverse=True,
        )

        total_count = len(leaves)
        if pagination.offset >= total_count:
            return PaginatedListResponse(stage_runs=[], total_count=total_count)
        return PaginatedListResponse(
            stage_runs=leaves[pagination.offset : pagination.limit],
            total_count=total_count,
        )

    def find_past(
        self, filter: GetStageRunByQueryFilter, pagination: Pagination
    ) -> PaginatedListResponse:
        filter_matches = self.find(filter)
        parent_ids = set(u.parent_id for u in self._stage_runs)

        past = sorted(
            [stage_run for stage_run in filter_matches if stage_run.id in parent_ids],
            key=lambda x: x.created_at,
            reverse=True,
        )

        total_count = len(past)
        if pagination.offset >= total_count:
            return PaginatedListResponse(stage_runs=[], total_count=total_count)

        return PaginatedListResponse(
            stage_runs=past[pagination.offset : pagination.limit],
            total_count=total_count,
        )

    def create_initial(self, stage: str, data: dict = {}) -> StageRun:
        stage_run = StageRun(
            id=str(uuid.uuid4()),
            status="waiting",
            stage=stage,
            data=data,
            parent_id=None,
            created_at=datetime.now(),
            updated_at=datetime.now(),
        )

        self._stage_runs.append(stage_run)

        return stage_run

    def create_detached(self, stage_id: str, thread_data: dict = {}) -> StageRun:
        stage_run = DetachedStageRun(
            id=str(uuid.uuid4()),
            status="waiting",
            stage=stage_id,
            data=thread_data,
            parent_id=None,
            created_at=datetime.now(),
            updated_at=datetime.now(),
        )

        self._stage_runs.append(stage_run)

        return stage_run

    def create_next(self, parent: StageRun, dtos: List[Dict]) -> List[StageRun]:
        created = []
        for dto in self.hydrate_next_dto(parent, dtos):
            stage_run = StageRun(
                id=str(uuid.uuid4()),
                parent_id=parent.id,
                status="waiting",
                **dto,
                created_at=datetime.now(),
                updated_at=datetime.now(),
            )
            created.append(stage_run)
            self._stage_runs.append(stage_run)
        return created

    def change_status(
        self, id: str, status: Status, execution_id: Optional[str] = None
    ) -> bool:
        stage_run = self.get(id)
        stage_run.updated_at = datetime.now()

        if not stage_run.execution_id and execution_id is not None:
            stage_run.execution_id = execution_id

        if stage_run.status in end_status:
            return False

        if status in status_transitions.get(stage_run.status, []):
            stage_run.status = status
            return True

        return False

    def fork(
        self,
        stage_run: StageRun,
        parent_is_iterator: bool,
        custom_thread_data: Optional[str] = None,
    ) -> StageRun:
        if stage_run.parent_id is None:
            return self.create_initial(stage_run.stage, {})
        parent_stage_run = self.get(stage_run.parent_id)
        new_stage_run = stage_run.clone_to_waiting()
        new_stage_run.data = parent_stage_run.data
        if parent_is_iterator:
            new_stage_run.data["item"] = stage_run.data["item"]
        new_stage_run_dto = new_stage_run.to_dto()
        stage_runs = self.create_next(parent_stage_run, [new_stage_run_dto])
        if custom_thread_data:
            validate_json(custom_thread_data)
            stage_runs[0].data = json.loads(custom_thread_data)
        return stage_runs[0]

    def get_first_retried_stage_run_initial_data(self, stage_run: StageRun) -> dict:
        if stage_run.parent_id is None:
            return {}
        parent = self.get(stage_run.parent_id)
        if parent.stage == stage_run.stage and parent.status == "failed":
            return self.get_first_retried_stage_run_initial_data(parent)
        return parent.data

    def retry(self, stage_run: StageRun) -> StageRun:
        if stage_run.status != "failed":
            raise Exception(
                f"Can't retry stage run {stage_run.id} with non failed status"
            )
        initial_data = self.get_first_retried_stage_run_initial_data(stage_run)
        retried_stage_run = stage_run.clone_to_waiting()
        retried_stage_run_dto = retried_stage_run.to_dto()
        stage_runs = self.create_next(stage_run, [retried_stage_run_dto])
        stage_runs[0].data = initial_data
        return stage_runs[0]

    def update_data(self, stage_run_id: str, data: dict) -> bool:
        stage_run = self.get(stage_run_id)
        stage_run.updated_at = datetime.now()
        stage_run.data = data
        return True


class RemoteStageRunRepository(StageRunRepository):
    def _request(
        self,
        method: str,
        path: str,
        body: Any = None,
        params: dict = {},
        raise_for_status: bool = True,
    ):
        headers: Mapping[str, str] = SIDECAR_HEADERS
        r = requests.request(
            method=method,
            url=f"{SIDECAR_URL}/stage-runs{path}",
            headers=headers,
            json=body,
            params=params,
        )

        if raise_for_status:
            r.raise_for_status()

        return r

    @staticmethod
    def create_from_dto(dto: dict) -> StageRun:
        return StageRun(
            id=dto["id"],
            stage=dto["stage"],
            data=dto["data"],
            status=dto["status"],
            parent_id=dto.get("parentId"),
            execution_id=dto.get("executionId"),
            created_at=from_utc_iso_string(dto["createdAt"]),
            updated_at=from_utc_iso_string(dto["updatedAt"]),
        )

    def get(self, id: str) -> StageRun:
        r = self._request("GET", path=f"/{id}")
        return self.create_from_dto(r.json())

    def find(
        self, filter: GetStageRunByQueryFilter, filter_detached: Optional[bool] = True
    ) -> List[StageRun]:
        r = self._request("GET", path="/", params=filter.to_dict())
        return [self.create_from_dto(dto) for dto in r.json()]

    def find_ancestors(self, id: str) -> List[StageRun]:
        r = self._request("GET", path=f"/{id}/ancestors")
        return [self.create_from_dto(dto) for dto in r.json()]

    def create_initial(self, stage: str, data: dict = {}) -> StageRun:
        body = dict(data={}, stage=stage)
        r = self._request("POST", path="/", body=body)
        return self.create_from_dto(r.json())

    def create_detached(self, stage_id: str, thread_data: dict = {}) -> StageRun:
        raise Exception("Attempt to create detached stage run in remote repository")

    def remove(self, id: str) -> None:
        raise Exception("Attempt to remove stage run in remote repository")

    def create_next(self, parent: StageRun, dtos: List[Dict]) -> List[StageRun]:
        if len(dtos) == 0:
            return []

        parent_id = parent.id
        next_dtos = self.hydrate_next_dto(parent, dtos)
        r = self._request("PUT", path=f"/{parent_id}/children", body=next_dtos)
        return [self.create_from_dto(dto) for dto in r.json()]

    def change_status(
        self, id: str, status: str, execution_id: Optional[str] = None
    ) -> bool:
        r = self._request(
            "PATCH", path=f"/{id}", body={"status": status}, raise_for_status=False
        )

        if r.status_code == 409:
            return False

        r.raise_for_status()

        return True

    def find_leaves(
        self,
        filter: GetStageRunByQueryFilter,
        pagination: Pagination,
    ) -> PaginatedListResponse:
        r = self._request(
            "GET",
            path="/leaves",
            params={**filter.to_dict(), **pagination.to_dict()},
            raise_for_status=False,
        )

        r.raise_for_status()

        return PaginatedListResponse.from_dict(r.json())

    def find_past(
        self, filter: GetStageRunByQueryFilter, pagination: Pagination
    ) -> PaginatedListResponse:
        raise NotImplementedError()

    def fork(
        self,
        stage_run: StageRun,
        parent_is_iterator: bool,
        custom_thread_data: Optional[str],
    ) -> StageRun:
        raise NotImplementedError()

    def retry(self, stage_run: StageRun) -> StageRun:
        if stage_run.status != "failed":
            raise Exception(
                f"Can't retry stage run {stage_run.id} with non failed status"
            )
        ancestors = self.find_ancestors(stage_run.id)
        initial_data = {}
        for ancestor in ancestors:
            if ancestor.status != "failed" or ancestor.stage != stage_run.stage:
                initial_data = ancestor.data
        retried_stage_run = stage_run.clone_to_waiting()
        retried_stage_run_dto = retried_stage_run.to_dto()
        stage_runs = self.create_next(stage_run, [retried_stage_run_dto])
        stage_runs[0].data = initial_data
        return stage_runs[0]

    def update_data(self, stage_run_id: str, data: dict) -> bool:
        r = self._request("PATCH", path=f"/{stage_run_id}", body={"data": data})
        if r.status_code == 409:
            return False

        r.raise_for_status()

        return True


def stage_run_repository_factory() -> StageRunRepository:
    if SIDECAR_URL is None:
        return LocalStageRunRepository()
    else:
        return RemoteStageRunRepository()



from .LinkOutput import LinkOutput
from .FileInput import FileInput
from .ProgressOutput import ProgressOutput
from .TagInput import TagInput
from .HtmlOutput import HtmlOutput
from .CnpjInput import CnpjInput
from .PlotlyOutput import PlotlyOutput
from .PhoneInput import PhoneInput
from .EmailInput import EmailInput
from .TextInput import TextInput
from .CheckboxInput import CheckboxInput
from .PandasRowSelectionInput import PandasRowSelectionInput
from .TimeInput import TimeInput
from .CustomInput import CustomInput
from .ListInput import ListInput
from .NumberInput import NumberInput
from .CardsInput import CardsInput
from .CurrencyInput import CurrencyInput
from .NpsInput import NpsInput
from .MultipleChoiceInput import MultipleChoiceInput
from .ChecklistInput import ChecklistInput
from .PasswordInput import PasswordInput
from .ImageOutput import ImageOutput
from .DateInput import DateInput
from .MarkdownOutput import MarkdownOutput
from .ImageInput import ImageInput
from .CodeInput import CodeInput
from .NumberSliderInput import NumberSliderInput
from .TextareaInput import TextareaInput
from .PandasOutput import PandasOutput
from .DropdownInput import DropdownInput
from .TextOutput import TextOutput
from .CpfInput import CpfInput
from .LatexOutput import LatexOutput
from .IframeOutput import IframeOutput
from .FileOutput import FileOutput
from .ToggleInput import ToggleInput
from .VideoInput import VideoInput
from .CameraInput import CameraInput
from .RatingInput import RatingInput
from .RichTextInput import RichTextInput



import abc
from typing import Dict, List, Literal, Optional, TypedDict

import flask_sock

from abstra_internals.compatibility import compat_traceback
from abstra_internals.contract import forms_contract
from abstra_internals.contract.forms import BrowserMessageTypes
from abstra_internals.entities.execution import RequestContext
from abstra_internals.utils import deserialize, serialize
from abstra_internals.utils.pthread_store import PThreadStore


class ClientAbandoned(Exception):
    pass


class Response(TypedDict):
    status: int
    body: str
    headers: Dict[str, str]


class ExecutionClient(abc.ABC):
    @abc.abstractmethod
    def handle_lock_failed(self, status: str) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def handle_failure(self, e: Exception) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def handle_success(self) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def handle_stdio(self, event: Literal["stdout", "stderr"], text: str) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def handle_start(self, execution_id: str):
        raise NotImplementedError()

    @abc.abstractmethod
    def handle_unset_thread(self):
        raise NotImplementedError()


class BasicClient(ExecutionClient):
    def handle_lock_failed(self, status: str) -> None:
        pass

    def handle_failure(self, e: Exception) -> None:
        pass

    def handle_success(self) -> None:
        pass

    def handle_stdio(self, event: Literal["stdout", "stderr"], text: str) -> None:
        pass

    def handle_start(self, execution_id: str):
        pass

    def handle_unset_thread(self):
        pass


class HookClient(ExecutionClient):
    def __init__(self, request_context: RequestContext) -> None:
        self.response = Response(status=200, body="", headers={})
        self.request_context = request_context

    def handle_lock_failed(self, status: str) -> None:
        self.response["status"] = 423
        pass

    def handle_failure(self, e: Exception) -> None:
        self.response["status"] = 500
        self.response["body"] = "An exception occurred during execution."
        pass

    def handle_success(self) -> None:
        pass

    def set_response(self, status: int, body: str, headers: Dict[str, str]) -> None:
        self.response["status"] = status
        self.response["body"] = body
        self.response["headers"] = headers

    def get_request(self) -> RequestContext:
        return self.request_context

    def handle_stdio(self, event: Literal["stdout", "stderr"], text: str) -> None:
        pass

    def handle_start(self, execution_id: str):
        pass

    def handle_unset_thread(self):
        self.response["status"] = 423


class FormClient(ExecutionClient):
    def __init__(self, ws: flask_sock.Server, request_context: RequestContext) -> None:
        self._ws = ws
        self._request_context = request_context

    def request_mount_page(
        self,
        widgets: list,
        actions: list,
        end_program: bool,
        reactive_polling_interval: int,
        steps_info: Optional[List],
    ) -> None:
        self._send(
            forms_contract.FormMountPageMessage(
                widgets,
                actions,
                end_program,
                reactive_polling_interval,
                steps_info,
            )
        )

    def request_page_update(
        self, widgets: list, validation: forms_contract.ValidationResult, event_seq: int
    ) -> None:
        self._send(forms_contract.FormUpdatePageMessage(widgets, validation, event_seq))

    def get_query_params(self) -> Dict[str, str]:
        return self._request_context.get("query_params", {})

    def request_auth(self, refresh: bool = False):
        self._send(forms_contract.AuthRequireInfoMessage(refresh=refresh))
        return self.wait_for_message("auth:saved-jwt")

    def handle_invalid_jwt(self):
        self._send(forms_contract.AuthInvalidJWTMessage())

    def handle_valid_jwt(self):
        self._send(forms_contract.AuthValidJWTMessage())

    def request_execute_js(self, code: str, context: dict = {}):
        self._send(forms_contract.ExecuteJSRequestMessage(code, context))
        data = self.wait_for_message("execute-js:response")
        return data.get("value")

    def request_redirect(
        self, url: str, query_params: Optional[Dict[str, str]]
    ) -> None:
        _query_params = (
            query_params if query_params is not None else self.get_query_params()
        )
        self._send(forms_contract.RedirectRequestMessage(url, _query_params))

    def handle_success(self):
        close_dto = forms_contract.CloseDTO(exit_status="SUCCESS")
        try:
            self._send(forms_contract.ExecutionEndedMessage(close_dto))
        except ClientAbandoned:
            pass

    def handle_start(self, execution_id: str):
        self.wait_for_message("execution:start")
        self._send(forms_contract.ExecutionStartedMessage(execution_id))

    def handle_unset_thread(self):
        self.handle_failure(Exception("Thread was unset"))

    def handle_failure(self, e: Exception):
        stdio_msg = forms_contract.ExecutionStdioMessage(
            "stderr",
            "".join(
                compat_traceback.format_exception(e),
            ),
        )
        try:
            self._send(stdio_msg)
            close_dto = forms_contract.CloseDTO(exit_status="EXCEPTION", exception=e)
            self._send(forms_contract.ExecutionEndedMessage(close_dto))
        except ClientAbandoned:
            pass

    def handle_lock_failed(self, status: str) -> None:
        self._send(forms_contract.ExecutionLockFailedMessage(status))

    def handle_stdio(self, event: Literal["stdout", "stderr"], text: str) -> None:
        self._send(forms_contract.ExecutionStdioMessage(event, text))

    def _receive_message(self):
        try:
            str_data = self._ws.receive()
            deserialized = deserialize(str_data)
            if not isinstance(deserialized, dict):
                raise ValueError("Invalid message format")
            return deserialized
        except flask_sock.ConnectionClosed:
            raise ClientAbandoned()

    def wait_for_message(
        self, target_type: Optional[BrowserMessageTypes] = None
    ) -> Dict:
        ignored_types = ["execution:heartbeat", "debug:close-attempt"]

        while True:
            message = self._receive_message()
            message_type = message.get("type")

            if message_type in ignored_types:
                continue
            if target_type and message_type != target_type:
                continue

            return message

    def _send(self, msg: forms_contract.Message) -> None:
        str_data = serialize(msg.to_json())
        try:
            self._ws.send(str_data)
        except flask_sock.ConnectionClosed:
            raise ClientAbandoned()


class ClientTypeMismatch(Exception):
    pass


class ClientNotFound(Exception):
    pass


class ExecutionClientStore:
    store: PThreadStore[ExecutionClient] = PThreadStore()

    @classmethod
    def set(cls, client: ExecutionClient):
        cls.store.set(client)

    @classmethod
    def get(cls) -> ExecutionClient:
        client = cls.store.get()
        if not client:
            raise ClientNotFound()
        return client

    @classmethod
    def get_hook_client(cls) -> HookClient:
        client = cls.get()
        if not isinstance(client, HookClient):
            raise ClientTypeMismatch
        return client

    @classmethod
    def get_form_client(cls) -> FormClient:
        client = cls.get()
        if not isinstance(client, FormClient):
            raise ClientTypeMismatch
        return client

    @classmethod
    def clear(cls):
        cls.store.clear()



from abc import ABC, abstractmethod
from typing import Dict, List, Optional

import requests

from abstra_internals.entities.execution import (
    ExecutionDTO,
    ExecutionStatus,
)
from abstra_internals.environment import (
    SERVER_UUID,
    SIDECAR_HEADERS,
    SIDECAR_URL,
    WORKER_UUID,
)
from abstra_internals.utils.pthread_store import PThreadStore


class ExecutionRepository(ABC):
    pthread_store: PThreadStore[ExecutionDTO]

    def get_current(self) -> Optional[ExecutionDTO]:
        return self.pthread_store.get()

    def free_current(self) -> None:
        self.pthread_store.clear()

    @abstractmethod
    def create(self, execution_dto: ExecutionDTO) -> None:
        raise NotImplementedError()

    @abstractmethod
    def update(self, execution_dto: ExecutionDTO) -> None:
        raise NotImplementedError()

    @abstractmethod
    def set_failure_by_id(self, execution_id: str) -> None:
        raise NotImplementedError()

    @abstractmethod
    def find_by_worker(
        self, app_id: str, worker_id: str, status: ExecutionStatus
    ) -> List[ExecutionDTO]:
        raise NotImplementedError()


class EditorExecutionRepository(ExecutionRepository):
    def __init__(self):
        self.pthread_store = PThreadStore()

    def create(self, execution_dto: ExecutionDTO) -> None:
        self.pthread_store.set(execution_dto)

    def update(self, execution_dto: ExecutionDTO) -> None:
        self.pthread_store.set(execution_dto)

    def set_failure_by_id(self, execution_id: str) -> None:
        execution = next(
            (
                execution
                for execution in self.pthread_store.items.values()
                if execution["id"] == execution_id
            ),
            None,
        )

        if execution:
            execution["status"] = "failed"

    def find_by_worker(
        self, app_id: str, worker_id: str, status: ExecutionStatus
    ) -> List[ExecutionDTO]:
        raise NotImplementedError()


class CloudExecutionRepository(ExecutionRepository):
    def __init__(
        self,
        url: str,
        headers: Dict[str, str],
    ):
        self.url = url
        self.headers = headers
        self.pthread_store = PThreadStore()

    def create(self, execution_dto: ExecutionDTO) -> None:
        self.pthread_store.set(execution_dto)

        request_dto = dict(
            id=execution_dto["id"],
            status=execution_dto["status"],
            createdAt=execution_dto["created_at"],
            context=execution_dto["context"] if execution_dto["context"] else {},
            stageId=execution_dto["stage_id"],
            stageRunId=execution_dto["stage_run_id"],
            workerId=WORKER_UUID(),
            appId=SERVER_UUID(),
        )

        res = requests.post(
            f"{self.url}/executions",
            json=request_dto,
            headers=self.headers,
        )

        res.raise_for_status()

    def update(self, execution_dto: ExecutionDTO) -> None:
        self.pthread_store.set(execution_dto)

        request_dto = dict(
            status=execution_dto["status"],
            context=execution_dto["context"] if execution_dto["context"] else {},
            stageRunId=execution_dto["stage_run_id"],
        )

        res = requests.patch(
            f"{self.url}/executions/{execution_dto['id']}",
            json=request_dto,
            headers=self.headers,
        )

        res.raise_for_status()

    def set_failure_by_id(self, execution_id: str) -> None:
        res = requests.patch(
            f"{self.url}/executions/{execution_id}",
            json=dict(status="failed"),
            headers=self.headers,
        )

        res.raise_for_status()

    def find_by_worker(
        self, app_id: str, worker_id: str, status: ExecutionStatus
    ) -> List[ExecutionDTO]:
        res = requests.get(
            f"{self.url}/executions",
            params=dict(
                appId=app_id,
                status=status,
                workerId=worker_id,
            ),
            headers=self.headers,
        )

        res.raise_for_status()

        return [
            ExecutionDTO(
                id=cloud_dto["id"],
                status=cloud_dto["status"],
                created_at=cloud_dto["createdAt"],
                stage_id=cloud_dto["stageId"],
                stage_run_id=cloud_dto["stageRunId"],
                context=cloud_dto["context"],
            )
            for cloud_dto in res.json()
        ]


def execution_repository_factory() -> ExecutionRepository:
    if SIDECAR_URL:
        return CloudExecutionRepository(
            url=SIDECAR_URL,
            headers=SIDECAR_HEADERS,
        )
    else:
        return EditorExecutionRepository()



from typing import Dict, List, Tuple, Union

from abstra_internals.controllers.sdk import HookSDKController
from abstra_internals.utils import serialize


def get_raw_request() -> Tuple[str, Dict[str, str], Dict[str, str]]:
    sdk_controller = HookSDKController()
    return sdk_controller.get_raw_request()


def get_request() -> Tuple[Union[str, List, Dict], Dict[str, str], Dict[str, str]]:
    sdk_controller = HookSDKController()
    return sdk_controller.get_request()


def send_response(body="", status_code=200, headers={}) -> None:
    sdk_controller = HookSDKController()
    sdk_controller.set_response(status_code, body, headers)


def send_json(data={}, status_code=200, headers={}):
    send_response(serialize(data), status_code, headers)



import unittest

from abstra_internals.utils.filter import (
    Condition,
    FilterCondition,
    LogicalGroup,
    evaluate_condition,
    evaluate_logical_group,
)


class TestEvaluateCondition(unittest.TestCase):
    def test_is_comparator(self):
        self.assertTrue(
            evaluate_condition(Condition("key", "is", "value"), {"key": "value"})
        )
        self.assertFalse(
            evaluate_condition(Condition("key", "is", "value"), {"key": "not_value"})
        )

    def test_contains_comparator_with_string(self):
        self.assertTrue(
            evaluate_condition(
                Condition("key", "contains", "value"), {"key": "This is a value"}
            )
        )
        self.assertFalse(
            evaluate_condition(
                Condition("key", "contains", "value"),
                {"key": "This does not contain the word."},
            )
        )

    def test_contains_comparator_with_not_string(self):
        self.assertFalse(
            evaluate_condition(Condition("key", "contains", "value"), {"key": 123})
        )

    def test_condition_with_missing_key(self):
        self.assertFalse(
            evaluate_condition(Condition("key", "is", "value"), {"other_key": "value"})
        )


class TestEvaluateLogicalGroup(unittest.TestCase):
    def test_and_group(self):
        group = LogicalGroup(
            AND=[Condition("key1", "is", "value1"), Condition("key2", "is", "value2")]
        )
        self.assertTrue(
            evaluate_logical_group(group, {"key1": "value1", "key2": "value2"})
        )
        self.assertFalse(
            evaluate_logical_group(group, {"key1": "value1", "key2": "not_value_2"})
        )

    def test_or_group(self):
        group = LogicalGroup(
            OR=[Condition("key1", "is", "value1"), Condition("key2", "is", "value2")]
        )
        self.assertTrue(
            evaluate_logical_group(group, {"key1": "value1", "key2": "value2"})
        )
        self.assertTrue(
            evaluate_logical_group(group, {"key1": "value1", "key2": "not_value_2"})
        )
        self.assertFalse(
            evaluate_logical_group(
                group, {"key1": "not_value_1", "key2": "not_value_2"}
            )
        )

    def test_not_group(self):
        group = LogicalGroup(NOT=Condition("key", "is", "value"))
        self.assertFalse(evaluate_logical_group(group, {"key": "value"}))
        self.assertTrue(evaluate_logical_group(group, {"key": "not_value"}))

    def test_invalid_group(self):
        group = LogicalGroup()
        self.assertFalse(evaluate_logical_group(group, {"key": "value"}))


class TestEvaluateFilterCondition(unittest.TestCase):
    def test_nested_condition(self):
        group = LogicalGroup(
            AND=[
                LogicalGroup(
                    OR=[
                        Condition("key1", "is", "value1"),
                        LogicalGroup(NOT=Condition("key2", "is", "value2")),
                    ]
                ),
                LogicalGroup(
                    OR=[
                        Condition("key2", "is", "value2"),
                        Condition("key3", "is", "value3"),
                    ]
                ),
            ]
        )
        self.assertTrue(
            evaluate_logical_group(
                group, {"key1": "value1", "key2": "not_value_2", "key3": "value3"}
            )
        )
        self.assertTrue(
            evaluate_logical_group(
                group, {"key1": "value1", "key2": "value2", "key3": "value3"}
            )
        )
        self.assertFalse(
            evaluate_logical_group(
                group, {"key1": "not_value_1", "key2": "value2", "key3": "value3"}
            )
        )

    def test_none_is_evaluated_as_true_except_for_not(self):
        group = LogicalGroup(AND=[None])
        self.assertTrue(evaluate_logical_group(group, {}))
        group = LogicalGroup(OR=[None])
        self.assertTrue(evaluate_logical_group(group, {}))
        group = LogicalGroup(NOT=None)
        self.assertFalse(evaluate_logical_group(group, {}))


class TestFilterCondition(unittest.TestCase):
    def test_from_dict_condition(self):
        data = {"key": "key", "comparator": "is", "value": "value"}
        condition = FilterCondition.from_dict(data)
        self.assertIsInstance(condition, Condition)
        self.assertEqual(condition, Condition("key", "is", "value"))

    def test_from_dict_complex_logic_group(self):
        data = {
            "AND": [
                {"NOT": {"key": "key1", "comparator": "is", "value": "value1"}},
                {"OR": [{"key": "key2", "comparator": "is", "value": "value2"}]},
            ]
        }
        group = FilterCondition.from_dict(data)
        self.assertIsInstance(group, LogicalGroup)
        self.assertEqual(
            group,
            LogicalGroup(
                AND=[
                    LogicalGroup(NOT=Condition("key1", "is", "value1")),
                    LogicalGroup(OR=[Condition("key2", "is", "value2")]),
                ]
            ),
        )

    def test_from_dict_invalid_data(self):
        data = {"invalid": "data"}
        self.assertIsNone(FilterCondition.from_dict(data))

        another_data = {"key": "key", "comparator": "is"}
        self.assertIsNone(FilterCondition.from_dict(another_data))



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Literal, Optional


@dataclass
class FilterCondition:
    @staticmethod
    def from_dict(data: dict) -> Optional[FilterCondition]:
        if "key" in data and "comparator" in data and "value" in data:
            return Condition(
                key=data["key"], comparator=data["comparator"], value=data["value"]
            )
        if "AND" in data:
            return LogicalGroup(
                AND=[FilterCondition.from_dict(condition) for condition in data["AND"]]
            )
        if "OR" in data:
            return LogicalGroup(
                OR=[FilterCondition.from_dict(condition) for condition in data["OR"]]
            )
        if "NOT" in data:
            return LogicalGroup(NOT=FilterCondition.from_dict(data["NOT"]))

        return None


@dataclass
class Condition(FilterCondition):
    key: str
    comparator: Literal["is", "contains"]
    value: str


@dataclass
class LogicalGroup(FilterCondition):
    AND: Optional[List[Optional[FilterCondition]]] = None
    OR: Optional[List[Optional[FilterCondition]]] = None
    NOT: Optional[FilterCondition] = None


def evaluate_condition(condition: Condition, data: dict):
    if condition.key not in data:
        return False

    if condition.comparator == "is":
        return data[condition.key] == condition.value

    if condition.comparator == "contains":
        if not isinstance(data[condition.key], str):
            return False
        return condition.value in data[condition.key]

    return False


def evaluate_logical_group(group: LogicalGroup, data: dict):
    if group.AND:
        return all(
            evaluate_filter_condition(condition, data) for condition in group.AND
        )
    if group.OR:
        return any(evaluate_filter_condition(condition, data) for condition in group.OR)
    if group.NOT:
        return not evaluate_filter_condition(group.NOT, data)

    return False


def evaluate_filter_condition(filter_condition: Optional[FilterCondition], data: dict):
    if isinstance(filter_condition, Condition):
        return evaluate_condition(filter_condition, data)
    if isinstance(filter_condition, LogicalGroup):
        return evaluate_logical_group(filter_condition, data)
    if filter_condition is None:
        return True

    return False



import json

import flask

from abstra_internals.controllers.execution import ExecutionController
from abstra_internals.controllers.execution_client import BasicClient
from abstra_internals.controllers.workflow import workflow_engine
from abstra_internals.repositories import (
    execution_logs_repository,
    execution_repository,
    stage_run_repository,
)
from abstra_internals.repositories.project.project import ProjectRepository
from abstra_internals.server.controller.main import MainController
from abstra_internals.usage import usage
from abstra_internals.utils import is_it_true


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_jobs", __name__)

    @bp.get("/<path:id>")
    @usage
    def _get_job(id: str):
        job = controller.get_job(id)
        if not job:
            flask.abort(404)
        return job.editor_dto

    @bp.get("/")
    @usage
    def _get_jobs():
        return [f.editor_dto for f in controller.get_jobs()]

    @bp.post("/")
    @usage
    def _create_job():
        data = flask.request.json
        if not data:
            flask.abort(400)
        title = data.get("title")
        file = data.get("file")
        if not title or not file:
            flask.abort(400)
        workflow_position = data.get("position", (0, 0))
        job = controller.create_job(title, file, workflow_position)
        return job.editor_dto

    @bp.put("/<path:id>")
    @usage
    def _update_stage(id: str):
        data = flask.request.json
        if not data:
            flask.abort(400)

        job = controller.update_stage(id, data)
        return job.editor_dto if job else None

    @bp.delete("/<path:id>")
    @usage
    def _delete_job(id: str):
        remove_file = flask.request.args.get(
            "remove_file", default=False, type=is_it_true
        )
        controller.delete_job(id, remove_file)
        return {"success": True}

    @bp.post("/<path:id>/run")
    @usage
    def _run_job(id: str):
        job = controller.get_job(id)
        if not job:
            flask.abort(404)

        execution_controller = ExecutionController(
            stage=job,
            stage_run_repository=stage_run_repository,
            target_stage_run_id=None,
            execution_repository=execution_repository,
            project_repository=ProjectRepository,
            request=None,
            client=BasicClient(),
        )

        execution_dto = execution_controller.run()

        if not execution_dto:
            return flask.abort(429)

        workflow_engine.handle_pthread_execution_end()

        return {"output": execution_logs_repository.get_logs_dto(execution_dto["id"])}

    @bp.post("/<path:id>/test")
    @usage
    def _test_job(id: str):
        job = controller.get_job(id)
        if not job:
            flask.abort(404)

        execution_controller = ExecutionController(
            stage=job,
            stage_run_repository=controller.stage_run_repository,
            target_stage_run_id=None,
            execution_repository=execution_repository,
            project_repository=ProjectRepository,
            request=None,
            client=BasicClient(),
        )

        thread_data = json.loads(controller.read_test_data())
        execution_dto = execution_controller.run_detached(thread_data=thread_data)

        if not execution_dto:
            return flask.abort(429)

        return {"output": execution_logs_repository.get_logs_dto(execution_dto["id"])}

    return bp



from unittest import TestCase

from abstra_internals.controllers.workflow import workflow_engine
from abstra_internals.repositories.project.project import (
    FormStage,
    JobStage,
    NotificationTrigger,
    WorkflowTransition,
)
from abstra_internals.repositories.stage_run import LocalStageRunRepository


class ActionTransitionsTest(TestCase):
    def setUp(self) -> None:
        self.repository = LocalStageRunRepository()

    def test_advances_when_finished(self):
        stage = JobStage(
            id="1",
            title="title",
            file="job.py",
            schedule="* * * * *",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="target_type",
                    target_id="target_id",
                    type="jobs:finished",
                )
            ],
        )
        stage_run = self.repository.create_initial(stage="1", data={})
        result = workflow_engine._follow_action_transitions(
            stage, stage_run, "jobs:finished"
        )
        self.assertEqual(result, [dict(stage="target_id", data={})])

    def test_stays_when_abandoned(self):
        stage = FormStage(
            id="1",
            title="title",
            file="job.py",
            path="path",
            is_initial=False,
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="target_type",
                    target_id="target_id",
                    type="jobs:finished",
                )
            ],
            notification_trigger=NotificationTrigger(
                variable_name="foo",
                enabled=False,
            ),
        )
        stage_run = self.repository.create_initial(stage="1", data={})
        result = workflow_engine._follow_action_transitions(
            stage, stage_run, "forms:abandoned"
        )
        self.assertEqual(result, [dict(stage="1", data={})])



import datetime
from typing import Any, Dict, Literal, Optional, TypedDict
from uuid import uuid4

from abstra_internals.utils.datetime import to_utc_iso_string

# Avoid circular import due to repo and entity being in the same file
ActionStage = Any


class RequestContext(TypedDict):
    body: str
    query_params: Dict[str, str]
    headers: Dict[str, str]
    method: str


ExecutionStatus = Literal["running", "failed", "finished", "abandoned"]


class ExecutionDTO(TypedDict):
    id: str
    status: ExecutionStatus
    created_at: str
    stage_id: str
    stage_run_id: str
    context: Optional[RequestContext]


class Execution:
    _status: ExecutionStatus = "running"
    id: str
    created_at: datetime.datetime
    stage: ActionStage
    stage_run_id: str
    request_context: Optional[RequestContext] = None

    def __init__(
        self,
        id: str,
        stage: ActionStage,
        stage_run_id: str,
        request_context: Optional[RequestContext],
        created_at: datetime.datetime,
    ) -> None:
        self.id = id
        self.created_at = created_at
        self.stage = stage
        self.stage_run_id = stage_run_id
        self.request_context = request_context

    @classmethod
    def create(
        cls,
        stage: ActionStage,
        stage_run_id: str,
        request_context: Optional[RequestContext],
    ):
        return cls(
            id=uuid4().__str__(),
            created_at=datetime.datetime.now(),
            stage=stage,
            stage_run_id=stage_run_id,
            request_context=request_context,
        )

    def set_status(self, status: ExecutionStatus) -> None:
        if status == "running":
            raise ValueError("Cannot set status to running")

        self._status = status

    def to_dto(self) -> ExecutionDTO:
        return ExecutionDTO(
            id=self.id,
            status=self._status,
            created_at=to_utc_iso_string(self.created_at),
            stage_id=self.stage.id,
            stage_run_id=self.stage_run_id,
            context=self.request_context,
        )



from pathlib import Path
from unittest import TestCase

from abstra_internals.controllers.execution import ExecutionController
from abstra_internals.controllers.execution_client import HookClient
from abstra_internals.entities.execution import RequestContext
from abstra_internals.repositories.execution import execution_repository_factory
from abstra_internals.repositories.project.project import (
    HookStage,
    ProjectRepository,
)
from abstra_internals.repositories.stage_run import (
    stage_run_repository_factory,
)
from tests.fixtures import clear_dir, init_dir


class ExecutionControllerTest(TestCase):
    def setUp(self) -> None:
        self.root = init_dir()
        self.stage_run_repository = stage_run_repository_factory()
        self.execution_repository = execution_repository_factory()
        self.request_context = RequestContext(
            body={"a": 1}.__str__(),
            headers={"auth": "some_secret_token"},
            query_params={"c": "3"},
            method="GET",
        )
        self.project = ProjectRepository.load()
        self.stage = HookStage.create(
            title="mock_stage",
            file="mock_file.py",
            id="mock_hook_id",
            workflow_position=(0, 1),
        )
        Path(self.stage.file).write_text("print('Hello, World!')")
        self.project.add_stage(self.stage)
        ProjectRepository.save(self.project)
        self.hook_client = HookClient(self.request_context)
        return super().setUp()

    def tearDown(self) -> None:
        clear_dir(self.root)
        return super().tearDown()

    def test_run_initial_returns_dto(self):
        controller = ExecutionController(
            stage=self.stage,
            target_stage_run_id=None,
            request=self.request_context,
            stage_run_repository=self.stage_run_repository,
            execution_repository=self.execution_repository,
            project_repository=ProjectRepository,
            client=self.hook_client,
        )

        execution_dto = controller.run()

        if not execution_dto:
            self.fail("ExecutionDTO is None")

        self.assertEqual(execution_dto["status"], "finished")

    def test_run_detached_returns_dto(self):
        controller = ExecutionController(
            stage=self.stage,
            target_stage_run_id=None,
            request=self.request_context,
            stage_run_repository=self.stage_run_repository,
            execution_repository=self.execution_repository,
            project_repository=ProjectRepository,
            client=self.hook_client,
        )

        execution_dto = controller.run_detached(thread_data={})

        if not execution_dto:
            self.fail("ExecutionDTO is None")

        self.assertEqual(execution_dto["status"], "finished")



import gc
from pathlib import Path
from typing import Literal, Tuple, Union

from abstra_internals.controllers.execution_client import (
    ClientAbandoned,
)
from abstra_internals.modules import import_as_new

Result = Tuple[Literal["finished", "abandoned", "failed"], Union[Exception, None]]


class ExecutionService:
    @staticmethod
    def _execute_code(filepath: Path):
        try:
            import_as_new(filepath.as_posix())
        except SystemExit as e:
            no_errors = e.code is None or e.code == 0
            if no_errors:
                return
            raise Exception(f"SystemExit: {e.code}") from e

    @classmethod
    def run(cls, filepath: Path) -> Result:
        try:
            cls._execute_code(filepath)
            return "finished", None
        except ClientAbandoned:
            return "abandoned", None
        except Exception as e:
            return "failed", e
        finally:
            gc.collect()



from typing import Literal

from abstra_internals.controllers.execution_client import (
    ClientNotFound,
    ExecutionClientStore,
)
from abstra_internals.repositories.execution import ExecutionRepository
from abstra_internals.repositories.execution_logs import ExecutionLogsRepository


class UnboundPthread(Exception):
    pass


class ExecutionStdioController:
    def __init__(
        self,
        execution_repository: ExecutionRepository,
        execution_logs_repository: ExecutionLogsRepository,
    ):
        self.execution_logs_repository = execution_logs_repository
        self.execution_repository = execution_repository

    def write_stdio(self, event: Literal["stderr", "stdout"], text: str) -> str:
        execution_dto = self.execution_repository.get_current()

        if not execution_dto:
            raise UnboundPthread()

        try:
            client = ExecutionClientStore.get()
        except ClientNotFound:
            raise UnboundPthread()

        client.handle_stdio(event, text)
        self.execution_logs_repository.insert_stdio(execution_dto["id"], event, text)

        return execution_dto["id"]



from abstra_internals.contract import forms as forms_contract
from abstra_internals.contract.forms import ExecutionStdioMessage, Message

__all__ = [
    "forms_contract",
    "Message",
    "ExecutionStdioMessage",
]



from pathlib import Path
from typing import Optional, Type

from abstra_internals.compatibility.compat_traceback import print_exception
from abstra_internals.controllers.execution_client import (
    ExecutionClient,
    ExecutionClientStore,
)
from abstra_internals.entities.execution import (
    Execution,
    ExecutionDTO,
    RequestContext,
)
from abstra_internals.repositories.execution import ExecutionRepository
from abstra_internals.repositories.project.project import (
    ActionStage,
    ProjectRepository,
)
from abstra_internals.repositories.stage_run import (
    StageRunRepository,
)
from abstra_internals.services.execution_service import ExecutionService

STAGE_RUN_ID_PARAM_KEY = "abstra-run-id"


class UnsetThread(Exception):
    pass


class ExecutionController:
    def __init__(
        self,
        stage: ActionStage,
        target_stage_run_id: Optional[str],
        request: Optional[RequestContext],
        stage_run_repository: StageRunRepository,
        execution_repository: ExecutionRepository,
        project_repository: Type[ProjectRepository],
        client: ExecutionClient,
    ) -> None:
        self.stage_run_repo = stage_run_repository
        self.execution_repo = execution_repository
        self.project_repo = project_repository

        self.client = client
        self.stage = stage
        self.target_stage_run_id = target_stage_run_id
        self.request = request

    def _should_create_initial_stage_run(self) -> bool:
        # TODO: move logic to stage entity
        project = self.project_repo.load()
        if not project:
            raise Exception("Project not found")

        return project.is_initial(self.stage) and not self.target_stage_run_id

    def run_detached(self, thread_data: dict) -> Optional[ExecutionDTO]:
        test_stage_run = self.stage_run_repo.create_detached(
            self.stage.id, thread_data=thread_data
        )

        self.target_stage_run_id = test_stage_run.id
        dto = self.run()
        self.stage_run_repo.remove(test_stage_run.id)
        return dto

    def run(self) -> Optional[ExecutionDTO]:
        ExecutionClientStore.set(self.client)
        stage_run_id = self.target_stage_run_id

        if self._should_create_initial_stage_run():
            stage_run = self.stage_run_repo.create_initial(self.stage.id)
            stage_run_id = stage_run.id

        if not stage_run_id:
            self.client.handle_unset_thread()
            ExecutionClientStore.clear()
            return

        execution = Execution.create(
            stage=self.stage,
            request_context=self.request,
            stage_run_id=stage_run_id,
        )

        if self.target_stage_run_id and not self.stage_run_repo.acquire_lock(
            self.target_stage_run_id, execution.id
        ):
            self.client.handle_lock_failed(self.target_stage_run_id)
            ExecutionClientStore.clear()
            return

        self.execution_repo.create(execution.to_dto())
        self.client.handle_start(execution.id)

        status, exception = ExecutionService.run(Path(execution.stage.file))
        self.stage_run_repo.change_status(execution.stage_run_id, status)

        execution.set_status(status)
        self.execution_repo.update(execution.to_dto())

        if exception:
            print_exception(exception)
            self.client.handle_failure(exception)
        else:
            self.client.handle_success()

        ExecutionClientStore.clear()
        return execution.to_dto()
